APPLICABLE_OPTIMIZER_TYPE_LIST = ["adam", "sgd"]
APPLICABLE_VAL_METRIC_LIST = ["val/loss", "val/acc"]

DEFAULT_BATCH_SIZE = 256
DEFAULT_LR = 1e-4
DEFAULT_N_LAYERS = 2
DEFAULT_DROPOUT_P = 0
DEFAULT_BN = True
DEFAULT_L1_STRENGTH = 0
DEFAULT_WEIGHT_DECAY = 0
DEFAULT_OPTIMIZER_TYPE = "adam"

DEFAULT_RANDOM_SEED_NUMBER = 2022

DEFAULT_MAX_EPOCHS = 100
DEFAULT_AUTO_LR_FIND = False
